VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Directory"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit


' Klasse zum Verzeichnisdurchforsten
'
' Inhalt von einem Verzeichnis wird bei Abfrage
' einer der Eigenschaften der Klasse, mit Ausnahme
' von "Name", ausgelesen.
'
' Arne Elster 2009


Private Declare Function FindFirstFile Lib "kernel32" _
Alias "FindFirstFileA" ( _
    ByVal lpFileName As String, _
    lpFindFileData As WIN32_FIND_DATA _
) As Long

Private Declare Function FindNextFile Lib "kernel32" _
Alias "FindNextFileA" ( _
    ByVal hFindFile As Long, _
    lpFindFileData As WIN32_FIND_DATA _
) As Long

Private Declare Function FindClose Lib "kernel32" ( _
    ByVal hFindFile As Long _
) As Long

Private Declare Function FileTimeToSystemTime Lib "kernel32" ( _
    lpFileTime As FILETIME, lpSystemTime As SYSTEMTIME _
) As Long

Private Const MAX_PATH                  As Long = 260
Private Const MAXDWORD                  As Long = &HFFFF
Private Const INVALID_HANDLE_VALUE      As Long = -1
Private Const FILE_ATTRIBUTE_ARCHIVE    As Long = &H20
Private Const FILE_ATTRIBUTE_DIRECTORY  As Long = &H10
Private Const FILE_ATTRIBUTE_HIDDEN     As Long = &H2
Private Const FILE_ATTRIBUTE_NORMAL     As Long = &H80
Private Const FILE_ATTRIBUTE_READONLY   As Long = &H1
Private Const FILE_ATTRIBUTE_SYSTEM     As Long = &H4
Private Const FILE_ATTRIBUTE_TEMPORARY  As Long = &H100

Private Type SYSTEMTIME
    wYear                   As Integer
    wMonth                  As Integer
    wDayOfWeek              As Integer
    wDay                    As Integer
    wHour                   As Integer
    wMinute                 As Integer
    wSecond                 As Integer
    wMilliseconds           As Integer
End Type

Private Type FILETIME
    dwLowDateTime           As Long
    dwHighDateTime          As Long
End Type

Private Type WIN32_FIND_DATA
    dwFileAttributes        As Long
    FTCreationTime          As FILETIME
    FTLastAccessTime        As FILETIME
    FTLastWriteTime         As FILETIME
    nFileSizeHigh           As Long
    nFileSizeLow            As Long
    dwReserved0             As Long
    dwReserved1             As Long
    cFileName               As String * MAX_PATH
    cAlternate              As String * 14
End Type

Private m_strDirectory      As String
Private m_blnGotListing     As Boolean

Private m_clsSubDirs()      As Directory
Private m_lngSubDirCount    As Long
Private m_clsFiles()        As File
Private m_lngFileCount      As Long

Public Property Get name() As String
    name = m_strDirectory
End Property

Public Property Let name(ByVal strDir As String)
    If Not Right$(strDir, 1) = "\" Then
        m_strDirectory = strDir & "\"
        If Not DirectoryExists(m_strDirectory) Then Err.Raise 53
    Else
        m_strDirectory = strDir
    End If
    
    m_blnGotListing = False
End Property

Public Property Get FileCount() As Long
    If Not m_blnGotListing Then GetContent
    FileCount = m_lngFileCount
End Property

Public Property Get SubDirCount() As Long
    If Not m_blnGotListing Then GetContent
    SubDirCount = m_lngSubDirCount
End Property

Public Property Get SubDirectory(ByVal idx As Long) As Directory
    If Not m_blnGotListing Then GetContent
    Set SubDirectory = m_clsSubDirs(idx)
End Property

Public Property Get File(ByVal idx As Long) As File
    If Not m_blnGotListing Then GetContent
    Set File = m_clsFiles(idx)
End Property

Private Sub GetContent(Optional ByVal strFilter As String = "*.*")
    Dim strFilters()    As String
    Dim strTemp         As String
    Dim hSearch         As Long
    Dim i               As Long
    Dim udtFindData     As WIN32_FIND_DATA
    
    strFilters = Split(strFilter, ";")
    m_lngSubDirCount = 0
    m_lngFileCount = 0
    
    hSearch = FindFirstFile(m_strDirectory & "*", udtFindData)
    If hSearch <> INVALID_HANDLE_VALUE Then
        Do
            ' cFileName mit Null-Chars bis MAX_PATH aufgefllt, entfernen
            strTemp = StripNulls(udtFindData.cFileName)
            
            ' Referenzen auf dieses und berverzeichnis ignorieren
            If (strTemp <> ".") And (strTemp <> "..") Then
                ' Verzeichnisse und Dateien getrennt behandeln
                If (udtFindData.dwFileAttributes And FILE_ATTRIBUTE_DIRECTORY) Then
                    Dim clsDir As Directory
                    Set clsDir = New Directory
                    clsDir.name = m_strDirectory & strTemp
                    
                    ReDim Preserve m_clsSubDirs(m_lngSubDirCount)
                    Set m_clsSubDirs(m_lngSubDirCount) = clsDir
                    m_lngSubDirCount = m_lngSubDirCount + 1
                Else
                    ' Prfen ob Datei im Suchraster (bestimmt durch Filter) ist
                    For i = 0 To UBound(strFilters)
                        If strTemp Like strFilters(i) Then
                            Dim clsFile As File
                            Set clsFile = New File
                            
                            With clsFile
                                .name = strTemp
                                .Path = m_strDirectory
                                .size = To64BitInt(udtFindData.nFileSizeLow, _
                                                   udtFindData.nFileSizeHigh, _
                                                   False)
                            
                                .CreationDate = FormatAPIDate(udtFindData.FTCreationTime)
                                .ModifiedDate = FormatAPIDate(udtFindData.FTLastWriteTime)
                            End With
                            
                            ReDim Preserve m_clsFiles(m_lngFileCount)
                            Set m_clsFiles(m_lngFileCount) = clsFile
                            m_lngFileCount = m_lngFileCount + 1
                            Exit For
                        End If
                    Next
                End If
            End If
        ' FindNextFile gibt Wert <> 0 zurck fr gefundenen Eintrag
        Loop While FindNextFile(hSearch, udtFindData)
    End If

    FindClose hSearch
    m_blnGotListing = True
End Sub

' Bringt API Zeitformat in VB Form
Private Function FormatAPIDate(dt As FILETIME) As Date
    Dim udtSysTime As SYSTEMTIME
    
    If FileTimeToSystemTime(dt, udtSysTime) Then
        With udtSysTime
            FormatAPIDate = CDate(format$(.wDay, "00") & "." & _
                                  format$(.wMonth, "00") & "." & _
                                  .wYear & " " & _
                                  format$(.wHour, "00") & ":" & _
                                  format$(.wMinute, "00") & ":" & _
                                  format$(.wSecond, "00"))
        End With
    End If
End Function

' Entfernt Null-Chars am Ende eines Strings
Private Function StripNulls(OriginalStr As String) As String
    If (InStr(OriginalStr, ChrW$(0)) > 0) Then
        StripNulls = Left$(OriginalStr, InStr(OriginalStr, ChrW$(0)) - 1)
    Else
        StripNulls = OriginalStr
    End If
End Function

' Prft die Existenz eines Verzeichnisses auf einem Datentrger
Private Function DirectoryExists(strDir As String) As Boolean
    On Error GoTo ErrHandler
    DirectoryExists = GetAttr(strDir) And vbDirectory
ErrHandler:
End Function

' Verrechnet Lo und Hi Anteil eines 64 Bit Ints zu einem Decimal
Private Function To64BitInt( _
    ByVal lngLo As Long, _
    ByVal lngHi As Long, _
    Optional ByVal signed As Boolean = True _
) As Variant

    Dim value   As Variant

    If (lngLo And &H80000000) Then
        value = CDec(lngLo And &H7FFFFFFF) + CDec(2147483648#)
    Else
        value = CDec(lngLo)
    End If
    
    If (Not signed) And (lngHi And &H80000000) Then
        lngHi = lngHi And &H7FFFFFFF
        value = value + (CDec(lngHi) + CDec(2147483648#)) * CDec(4294967296#)
    Else
        value = value + CDec(lngHi) * CDec(4294967296#)
    End If
    
    To64BitInt = value
End Function
