Attribute VB_Name = "FFT"
Option Explicit


' Eigene Fast Fourier Transformation
'
' Arne Elster 2008


Public Const PI2    As Single = 6.28318530717959
Public Const LOG2   As Single = 0.693147180559945

Public Sub FFT(sngRealInp() As Single, sngImagInp() As Single, _
               sngRealOut() As Single, sngImagOut() As Single, _
               ByVal n As Long, ByVal lngBits As Long)

    Dim lngSize As Long

    Dim i As Long, j As Long, l As Long
    
    FFTReorder sngRealInp, sngImagInp, _
               sngRealOut, sngImagOut, _
               n, lngBits
    
    lngSize = 2&
    For i = 0 To lngBits - 1&
        For j = 0 To n - 1& Step lngSize
            Dim k1 As Long: k1 = j + lngSize \ 2&
            Dim k2 As Long: k2 = j
            
            For l = 0 To lngSize \ 2 - 1
                Dim rad As Single: rad = (PI2 * l) / lngSize
                Dim WRe As Single: WRe = Cos(rad)
                Dim WIm As Single: WIm = -Sin(rad)
                
                Dim real As Single: real = sngRealOut(k1) * WRe - sngImagOut(k1) * WIm
                Dim imag As Single: imag = sngRealOut(k1) * WIm + sngImagOut(k1) * WRe
                
                sngRealOut(k1) = sngRealOut(k2) - real
                sngImagOut(k1) = sngImagOut(k2) - imag
                sngRealOut(k2) = sngRealOut(k2) + real
                sngImagOut(k2) = sngImagOut(k2) + imag
                
                k1 = k1 + 1&
                k2 = k2 + 1&
            Next
        Next
        lngSize = lngSize * 2&
    Next
End Sub

Private Sub FFTReorder(sngRealInp() As Single, sngImagInp() As Single, _
                       sngRealOut() As Single, sngImagOut() As Single, _
                       ByVal n As Long, ByVal Bits As Long)

    Dim i As Long, l As Long
    
    For i = 0 To n - 1
        Dim k As Long: k = i
        Dim j As Long
        
        l = 0&
        For j = 0 To Bits - 1&
            l = (l * 2&) Or (k And 1&)
            k = k \ 2&
        Next
        
        sngRealOut(l) = sngRealInp(i)
        sngImagOut(l) = sngImagInp(i)
    Next
End Sub
