Attribute VB_Name = "ASMMemory"
Option Explicit

Private Declare Function VirtualAlloc Lib "kernel32" ( _
    ByVal lpAddress As Long, ByVal dwSize As Long, _
    ByVal flAllocType As Long, ByVal flProtect As Long _
) As Long

Private Declare Function VirtualFree Lib "kernel32" ( _
    ByVal lpAddress As Long, ByVal dwSize As Long, _
    ByVal dwFreeType As Long _
) As Long

Public Declare Function VirtualProtect Lib "kernel32" ( _
    ByVal lpAddress As Long, ByVal dwSize As Long, _
    ByVal flNewProtect As Long, lpflOldProtect As Long _
) As Long

Private Declare Sub CpyMem Lib "kernel32" _
Alias "RtlMoveMemory" ( _
    pDst As Any, pSrc As Any, Optional ByVal dwLen As Long = 4 _
)

Public Declare Function CallWindowProc Lib "user32" _
Alias "CallWindowProcA" ( _
    ptr As Any, p1 As Any, p2 As Any, p3 As Any, p4 As Any _
) As Long

Private Declare Function HeapAlloc Lib "kernel32" ( _
    ByVal hHeap As Long, _
    ByVal dwFlags As Long, _
    ByVal dwBytes As Long _
) As Long

Private Declare Function GetProcessHeap Lib "kernel32" () As Long

Private Declare Function HeapFree Lib "kernel32" (ByVal hHeap As Long, ByVal dwFlags As Long, lpMem As Any) As Long

Private Const HEAP_ZERO_MEMORY As Long = &H8
Private Const HEAP_GENERATE_EXCEPTIONS As Long = &H4

Private Enum VirtualFreeTypes
    MEM_DECOMMIT = &H4000
    MEM_RELEASE = &H8000
End Enum

Private Enum VirtualAllocTypes
    MEM_COMMIT = &H1000
    MEM_RESERVE = &H2000
    MEM_RESET = &H8000
    MEM_LARGE_PAGES = &H20000000
    MEM_PHYSICAL = &H100000
    MEM_WRITE_WATCH = &H200000
End Enum

Private Enum VirtualAllocPageFlags
    PAGE_EXECUTE = &H10
    PAGE_EXECUTE_READ = &H20
    PAGE_EXECUTE_READWRITE = &H40
    PAGE_EXECUTE_WRITECOPY = &H80
    PAGE_NOACCESS = &H1
    PAGE_READONLY = &H2
    PAGE_READWRITE = &H4
    PAGE_WRITECOPY = &H8
    PAGE_GUARD = &H100
    PAGE_NOCACHE = &H200
    PAGE_WRITECOMBINE = &H400
End Enum

Public Type Memory
    address                     As Long
    Bytes                       As Long
    InitPtr                     As Long
    MainPtr                     As Long
    DeInitPtr                   As Long
    DSPPtr                      As Long
End Type

Private Const IDE_ADDROF_REL    As Long = 22


Public Function ASMAllocMemory(ByVal lngBytes As Long) As Long
    ASMAllocMemory = HeapAlloc(GetProcessHeap(), HEAP_GENERATE_EXCEPTIONS Or HEAP_ZERO_MEMORY, lngBytes)
End Function

Public Sub ASMFreeMemory(ByVal lngPtr As Long)
    HeapFree GetProcessHeap(), 0, ByVal lngPtr
End Sub


Public Function AsmToMem(ByVal asm As String) As Memory
    Dim clsAsm   As ASMBler
    Dim udtMem   As Memory
    Dim btAsm()  As Byte
    
    Set clsAsm = New ASMBler
    
    ' erstmal nur Gre der Ausgabe feststellen
    ' weil base address noch nicht bekannt ist
    If Not clsAsm.Assemble(asm, True) Then
        Err.Raise 54321, , "(Zeile " & clsAsm.LastErrorLine & ")  " & clsAsm.LastErrorMessage
    End If
    
    udtMem = AllocMemory(clsAsm.OutputSize, , PAGE_EXECUTE_READWRITE)
    
    clsAsm.BaseAddress = udtMem.address
    If Not clsAsm.Assemble(asm) Then
        Err.Raise 54321, , "(Zeile " & clsAsm.LastErrorLine & ")  " & clsAsm.LastErrorMessage
    End If
    
    btAsm = clsAsm.GetOutput()
    CpyMem ByVal udtMem.address, btAsm(0), udtMem.Bytes
    udtMem.InitPtr = clsAsm.LabelAddress("global_init", vbTextCompare)
    udtMem.DeInitPtr = clsAsm.LabelAddress("global_free", vbTextCompare)
    udtMem.MainPtr = clsAsm.LabelAddress("fnc_main", vbTextCompare)
    udtMem.DSPPtr = clsAsm.LabelAddress("fnc_dsp", vbTextCompare)
    
    AsmToMem = udtMem
End Function


Public Function AsmStringToArray(ByVal asm As String) As Byte()
    Dim i       As Long
    Dim j       As Long
    Dim plen    As Long
    Dim bt()    As Byte
    
    ReDim bt(Len(asm) \ 2 - 1) As Byte
    
    For i = 1 To Len(asm) Step 2
        bt(j) = CLng("&H" & Mid$(asm, i, 2))
        j = j + 1
    Next
    
    AsmStringToArray = bt
End Function


Private Function AllocMemory( _
    ByVal Bytes As Long, _
    Optional ByVal lpAddr As Long = 0, _
    Optional ByVal PageFlags As VirtualAllocPageFlags = PAGE_READWRITE _
) As Memory

    With AllocMemory
        .address = VirtualAlloc(lpAddr, Bytes, MEM_COMMIT, PageFlags)
        .Bytes = Bytes
    End With
End Function


Public Function FreeMemory(udtMem As Memory) As Boolean
    VirtualFree udtMem.address, udtMem.Bytes, MEM_DECOMMIT

    udtMem.address = 0
    udtMem.Bytes = 0
End Function


