Attribute VB_Name = "OptimizeTree"
Option Explicit


Public Sub OptimizeNode(ByVal nd As ExpressionNode)
    Dim i As Long
    
    For i = 0 To nd.SubNodeCount - 1
        OptimizeNode nd.SubNode(i)
    Next
    
    Select Case nd.NodeType
        Case NODE_CONV_BYTE_TO_FLOAT, NODE_CONV_BYTE_TO_INT, _
             NODE_CONV_BYTE_TO_SHORT, NODE_CONV_INT_TO_FLOAT, _
             NODE_CONV_SHORT_TO_FLOAT, NODE_CONV_SHORT_TO_INT:
            OptimizeNodeConversion nd
            
        Case NODE_OP_ADD, NODE_OP_SUB, NODE_OP_MUL, NODE_OP_DIV, _
             NODE_OP_BAND, NODE_OP_BOR, NODE_OP_BXOR, NODE_OP_MOD, _
             NODE_OP_NEG:
            OptimizeNodeMath nd
            
    End Select
End Sub


Private Sub OptimizeNodeMath(ByVal nd As ExpressionNode)
    If nd.NodeType = NODE_OP_NEG Then
        If nd.SubNode(0).NodeType = NODE_VAL_VALUE Then
            nd.NodeType = NODE_VAL_VALUE
            nd.Value = -nd.SubNode(0).Value
            nd.RemSubNode nd.SubNode(0)
            Exit Sub
        End If
    End If
    
    If nd.SubNode(0).NodeType = NODE_VAL_VALUE Then
        If nd.SubNode(1).NodeType = NODE_VAL_VALUE Then
            Select Case nd.NodeType
                Case NODE_OP_ADD:
                    nd.Value = nd.SubNode(0).Value + nd.SubNode(1).Value
                Case NODE_OP_SUB:
                    nd.Value = nd.SubNode(0).Value - nd.SubNode(1).Value
                Case NODE_OP_MUL:
                    nd.Value = nd.SubNode(0).Value * nd.SubNode(1).Value
                Case NODE_OP_DIV:
                    nd.Value = nd.SubNode(0).Value / nd.SubNode(1).Value
            End Select
            nd.NodeType = NODE_VAL_VALUE
            nd.RemSubNode nd.SubNode(0)
            nd.RemSubNode nd.SubNode(0)
        End If
    End If
End Sub


' Aus (.. <= CONVERT_X_TO_Y <= VALUE) einfach (.. <= VALUE) machen
Private Sub OptimizeNodeConversion(ByVal nd As ExpressionNode)
    If nd.SubNode(0).NodeType = NODE_VAL_VALUE Then
        Select Case nd.NodeType
            Case NODE_CONV_BYTE_TO_FLOAT, NODE_CONV_INT_TO_FLOAT, _
                 NODE_CONV_SHORT_TO_FLOAT:
                nd.DataType = TYPE_FLOAT
                nd.Value = CSng(nd.SubNode(0).Value)
            Case NODE_CONV_BYTE_TO_INT, NODE_CONV_SHORT_TO_INT:
                nd.DataType = TYPE_INT
                nd.Value = CLng(nd.SubNode(0).Value)
            Case NODE_CONV_BYTE_TO_SHORT:
                nd.DataType = TYPE_SHORT
                nd.Value = CInt(nd.SubNode(0).Value)
        End Select
        nd.NodeType = NODE_VAL_VALUE
        nd.RemSubNode nd.SubNode(0)
    End If
End Sub
