VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Parser"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit


' Einfacher Basic Parser
'
' Beispiel fr Sprache:
' Sub Main()
'   Dim x As Int = Faculty(5)
'   Return
' End
'
' Function Faculty(a As Int) As Int
'   Dim i As Int = 0
'   Dim c As Int = 1
'
'   For i = 2 To a
'     c = c * i
'   End
'
'   Return c
' End
'
' Arne Elster 2009


Private Enum NAME_TYPE
    NAME_VARIABLE
    NAME_FUNCTION
End Enum

Private Type VarInfo
    Name                As String
    DataType            As DATA_TYPE
    Reference           As Boolean
    IsArray             As Boolean
    Size                As ExpressionNode
End Type

Private Type OperatorTypes
    Op                  As NODE_TYPE
    Types               As DATA_TYPE
End Type

Private Type ExpressionLookupItem
    TokenType           As TokenType
    NodeType            As NODE_TYPE
End Type

Private Type ExpressionLookup
    Items()             As ExpressionLookupItem
    Count               As Long
End Type

Private Type FunctionParam
    Name                As String
    DataType            As DATA_TYPE
    IsArray             As Boolean
    Reference           As Boolean
End Type

Private Type FunctionHead
    Name                As String
    DataType            As DATA_TYPE
    Params()            As FunctionParam
    ParamCount          As Long
End Type

Private Type RegistryName
    Name                As String
    NameType            As NAME_TYPE
    DataType            As DATA_TYPE
    IsRef               As Boolean
    IsParam             As Boolean
    IsFixedArray        As Boolean
End Type

Private Type RegistryLevel
    Names()             As RegistryName
    NameCount           As Long
End Type

Private m_clsScanner    As Scanner
Private m_colErrors     As Collection

Private m_udtRegNames() As RegistryLevel
Private m_lngRegLevels  As Long

Private m_udtExprLk()   As ExpressionLookup
Private m_lngExprLkLvs  As Long

Private m_udtOpTypes()  As OperatorTypes
Private m_lngOpTypCnt   As Long

Private m_udeImplConv() As DATA_TYPE
Private m_lngImplConvCnt As Long

Private m_fncCurrent    As ParserFunction

Private m_udtFunctions() As FunctionHead
Private m_lngFncCount   As Long

Private Sub Class_Initialize()
    Set m_clsScanner = New Scanner
    Set m_colErrors = New Collection
    InitExpressionLookup
    InitOpTypes
End Sub


'#################################################
'#################################################
' PARSER


Private Function ParseDataTypeStr(ByVal strType As String) As DATA_TYPE
    Select Case LCase$(strType)
        Case "int":     ParseDataTypeStr = TYPE_INT
        Case "short":   ParseDataTypeStr = TYPE_SHORT
        Case "byte":    ParseDataTypeStr = TYPE_BYTE
        Case "float":   ParseDataTypeStr = TYPE_FLOAT
        Case "bool":    ParseDataTypeStr = TYPE_BOOL
        Case Else:      ParseDataTypeStr = TYPE_UNKNOWN
    End Select
End Function


Public Function Parse( _
    ByVal strSource As String, _
    ByRef clsGlobal As ParserGlobal _
) As Boolean

    Set clsGlobal = New ParserGlobal

    ClearErrors
    ClearRegistry
    NameRegistryGrow

    ' Alle Funktionen/Subs finden
    ' damit man keine Prototypes braucht
    m_clsScanner.Scan strSource
    FindFunctionNames
    m_clsScanner.Reset
    
    If Match(TokenStreamBegin) Then
        ParseGlobal clsGlobal
        Match TokenStreamEnd
    End If
    
    Parse = (m_colErrors.Count = 0)
End Function

Private Sub FindFunctionNames()
    Dim udtHead As FunctionHead
    
    Do
        Select Case m_clsScanner.TokenID
            Case TokenProcedure, TokenFunction:
                Erase udtHead.Params
                udtHead.ParamCount = 0
                
                ParseFunctionHead udtHead, False
                
                If Len(udtHead.Name) > 0 Then
                    If Not NameRegistryAdd(udtHead.Name, _
                                           NAME_FUNCTION, _
                                           udtHead.DataType, _
                                           False, False, False) Then
                        
                        AddError PARSER_NAME_AMBIGUOS, strMsg:=udtHead.Name & " ist zweideutig"
                    End If
                    
                    ReDim Preserve m_udtFunctions(m_lngFncCount)
                    m_udtFunctions(m_lngFncCount) = udtHead
                    m_lngFncCount = m_lngFncCount + 1
                End If
            Case Else:
                m_clsScanner.Consume
        End Select
    Loop While m_clsScanner.TokenID <> TokenStreamEnd
End Sub

' Global kann eine Variablendeklaration oder Funktion/Sub
' geschrieben werden, a la VB
Private Sub ParseGlobal(ByVal clsGlobal As ParserGlobal)
    Do
        Select Case m_clsScanner.TokenID
            Case TokenEndOfInstruction: m_clsScanner.Consume
            Case TokenDim:              ParseVarDeclare clsGlobal
            Case Else:                  Exit Do
        End Select
    Loop
    
    Do
        Select Case m_clsScanner.TokenID
            Case TokenEndOfInstruction: m_clsScanner.Consume
            Case TokenFunction:         ParseFunction clsGlobal
            Case TokenProcedure:        ParseFunction clsGlobal
            Case TokenDim:              AddError PARSER_UNEXPECTED_TOKEN, strMsg:="Globale Variablendeklarationen nur ber Methoden"
                                        ParseVarDeclare clsGlobal
            Case Else:                  AddError PARSER_UNEXPECTED_TOKEN
                                        m_clsScanner.Consume
        End Select
    Loop While m_clsScanner.TokenID <> TokenStreamEnd
End Sub

' Gltig sowohl fr Sub als auch fr Function
Private Sub ParseFunction(ByVal clsGlobal As ParserGlobal)
    Dim clsFunc As ParserFunction
    Dim clsCode As ParserCodeBlock
    Dim udtHead As FunctionHead
    Dim i       As Long
    
    Set clsFunc = New ParserFunction
    Set clsCode = New ParserCodeBlock
    
    ParseFunctionHead udtHead
    With udtHead
        clsFunc.Name = .Name
        clsFunc.DataType = .DataType
        For i = 0 To .ParamCount - 1
            With .Params(i)
                clsFunc.AddParam .Name, .DataType, .IsArray, .Reference
            End With
        Next
    End With
    
    NameRegistryGrow
    
    For i = 0 To udtHead.ParamCount - 1
        With udtHead.Params(i)
            If Not NameRegistryAdd(.Name, NAME_VARIABLE, .DataType, .Reference, True, False) Then
                AddError PARSER_NAME_AMBIGUOS, strMsg:=.Name & " ist zweideutig"
            End If
        End With
    Next
    
    Set m_fncCurrent = clsFunc
    ParseCodeBlock clsCode
    Set clsFunc.Code = clsCode
    clsGlobal.AddFunction clsFunc

    NameRegistryShrink
        
    Match TokenEnd
    Match TokenEndOfInstruction
End Sub


Private Sub ParseFunctionHead( _
    udtHead As FunctionHead, _
    Optional ByVal blnErrors As Boolean = True _
)

    Dim blnIsSub As Boolean
    
    If Match(TokenProcedure, False) Then
        blnIsSub = True
    ElseIf Match(TokenFunction, False) Then
        blnIsSub = False
    Else
        If blnErrors Then
            AddError PARSER_UNEXPECTED_TOKEN, strMsg:="Sub/Function erwartet"
        End If
    End If
    
    If m_clsScanner.TokenID = TokenName Then
        udtHead.Name = m_clsScanner.Consume()
    Else
        If blnErrors Then
            AddError PARSER_UNEXPECTED_TOKEN, strMsg:="Funktionsname erwartet"
        End If
    End If
    
    Match TokenParenthesisLeft, blnErrors
    ParseParamList udtHead, blnErrors
    Match TokenParenthesisRight, blnErrors
    
    If Not blnIsSub Then
        Match TokenAs, blnErrors
            
        If m_clsScanner.TokenID = TokenName Then
            udtHead.DataType = ParseDataTypeStr(m_clsScanner.Consume())
            If udtHead.DataType = TYPE_UNKNOWN Then
                If blnErrors Then
                    AddError PARSER_TYPE_UNKNOWN, strMsg:="Unbekannter Rckgabetyp"
                End If
            End If
        Else
            If blnErrors Then AddError PARSER_TYPE_UNKNOWN
        End If
    Else
        udtHead.DataType = TYPE_NONE
    End If
    
    Match TokenEndOfInstruction, blnErrors
End Sub


Private Sub ParseParamList(udtHead As FunctionHead, ByVal blnShowErr As Boolean)
    Dim udtVar As VarInfo
    
    Do While m_clsScanner.TokenID = TokenName Or _
             m_clsScanner.TokenID = TokenRef

        ParseVar udtVar, True, False
        With udtHead
            ReDim Preserve .Params(.ParamCount) As FunctionParam
        
'            If udtVar.IsArray And Not udtVar.Reference Then
'                If blnShowErr Then
'                    AddError PARSER_TYPE_MISMATCH, strMsg:="Arrays knnen nur als Referenz bergeben werden"
'                End If
'            End If
            
            With .Params(.ParamCount)
                .Name = udtVar.Name
                .DataType = udtVar.DataType
                .IsArray = udtVar.IsArray
                .Reference = udtVar.Reference
            End With
            
            .ParamCount = .ParamCount + 1
        End With
        
        If Not Match(TokenSeperator, False) Then Exit Do
    Loop
End Sub


Private Sub ParseCodeBlock(ByVal clsCode As ParserCodeBlock)
    Dim clsExpr As ExpressionNode
    
    NameRegistryGrow
    Do
        Select Case m_clsScanner.TokenID
            Case TokenEndOfInstruction: m_clsScanner.Consume
            Case TokenIf:               ParseStatementIf clsCode
            Case TokenDim:              ParseVarDeclare clsCode
            Case TokenReDim:            ParseStatementReDim clsCode
            Case TokenFor:              ParseStatementFor clsCode
            Case TokenWhile:            ParseStatementWhile clsCode
            Case TokenReturn:           ParseStatementReturn clsCode
            Case TokenName:             ParseExpression clsExpr, True
                                        clsCode.AddStatement clsExpr
                                        If clsExpr.NodeType <> NODE_OP_ASSIGN Then
                                            If clsExpr.NodeType <> NODE_OP_CALL Then
                                                AddError PARSER_ERR_UNKNOWN, strMsg:="Ausdruck muss Zuweisung sein"
                                            End If
                                        End If
            Case Else:                  Exit Do
        End Select
    Loop
    NameRegistryShrink
End Sub

Private Sub CheckName( _
    ByVal strName As String, _
    ByVal udeExpectedType As NAME_TYPE, _
    ByVal udeTypeMask As DATA_TYPE, _
    ByVal blnAllowArray As Boolean, _
    ByVal blnMustBeRefParam As Boolean _
)

    Dim hVar As Long: hVar = NameRegistryGetNameHandle(strName)
    Dim udtInfo As RegistryName: udtInfo = NameRegistryInfo(hVar)
    
    If udeExpectedType <> udtInfo.NameType Then
        If udeExpectedType = NAME_VARIABLE Then AddError PARSER_VAR_EXPECTED
        If udeExpectedType = NAME_FUNCTION Then AddError PARSER_INVALID_FUNCTION
    End If
    
    If (udeTypeMask And udtInfo.DataType) = 0 Then
        AddError PARSER_TYPE_MISMATCH
    Else
        If Not blnAllowArray Then
            If (udtInfo.DataType And TYPE_ARRAY) Then
                AddError PARSER_TYPE_MISMATCH
            End If
        End If
    End If
    
    If blnMustBeRefParam Then
        If udtInfo.IsParam Then
            If Not udtInfo.IsRef Then
                AddError PARSER_TYPE_MISMATCH, strMsg:="ReDim: " & strName & " muss Referenz sein"
            End If
        End If
    End If
End Sub

Private Sub ParseStatementReDim(ByVal clsCode As ParserCodeBlock)
    Dim clsReDim    As ParserStatementReDim
    Dim clsExpr     As ExpressionNode
    Dim strName     As String
    Dim hVar        As Long
    
    Set clsReDim = New ParserStatementReDim
    
    If Match(TokenReDim) Then
        If m_clsScanner.TokenID = TokenName Then
            strName = m_clsScanner.Consume()
            CheckName strName, NAME_VARIABLE, TYPE_ARRAY, True, True
        Else
            AddError PARSER_UNEXPECTED_TOKEN, strMsg:="ReDim: ID erwartet"
        End If
        
        Match TokenBracketLeft
        ParseExpression clsExpr, False
        CheckTypes clsExpr, TYPE_INT
        OptimizeNode clsExpr
        Match TokenBracketRight
        
        hVar = NameRegistryGetNameHandle(strName)
        With NameRegistryInfo(hVar)
            If .IsFixedArray Then AddError PARSER_TYPE_MISMATCH, strMsg:="ReDim fr Fixed Array nicht mglich"
            clsReDim.DataType = .DataType
        End With
        clsReDim.Identifier = strName
        Set clsReDim.Size = clsExpr
        clsCode.AddStatement clsReDim
    End If
End Sub

Private Sub ParseStatementFor(ByVal clsCode As ParserCodeBlock)
    Dim clsVar      As ExpressionNode
    Dim clsExprFrom As ExpressionNode
    Dim clsExprTo   As ExpressionNode
    Dim clsFor      As ParserStatementFor
    Dim clsForCode  As ParserCodeBlock
    
    If Match(TokenFor) Then
        Set clsFor = New ParserStatementFor
        Set clsForCode = New ParserCodeBlock
        
        ParseExpressionFactor clsVar
        If clsVar.NodeType <> NODE_VAL_VARIABLE Then
            AddError PARSER_UNEXPECTED_TOKEN, strMsg:="Zhlervariable erwartet"
        End If
        If (clsVar.DataType And (TYPE_INT Or TYPE_SHORT Or TYPE_BYTE)) = 0 Then
            AddError PARSER_TYPE_MISMATCH, strMsg:="Nur ganzzahlige Zhler fr For Loop erlaubt"
        End If
        If clsVar.DataType And TYPE_ARRAY Then AddError PARSER_TYPE_MISMATCH, strMsg:="For Zhler kann kein Array sein"
        
        Match TokenOpAssign
        ParseExpression clsExprFrom, False
        CheckTypes clsExprFrom, clsVar.DataType
        OptimizeNode clsExprFrom
        Match TokenTo
        ParseExpression clsExprTo, False
        CheckTypes clsExprTo, clsVar.DataType
        OptimizeNode clsExprTo
        Match TokenEndOfInstruction
        ParseCodeBlock clsForCode
        Match TokenEnd
        Match TokenEndOfInstruction
        
        Set clsFor.Code = clsForCode
        Set clsFor.ExprFrom = clsExprFrom
        Set clsFor.ExprTo = clsExprTo
        Set clsFor.Counter = clsVar
        
        clsCode.AddStatement clsFor
    End If
End Sub


Private Sub ParseStatementWhile(ByVal clsCode As ParserCodeBlock)
    Dim clsExpr     As ExpressionNode
    Dim clsWhile    As ParserStatementWhile
    Dim clsLoopCode As ParserCodeBlock
    
    If Match(TokenWhile) Then
        Set clsWhile = New ParserStatementWhile
        Set clsLoopCode = New ParserCodeBlock
        
        ParseExpression clsExpr, False
        CheckTypes clsExpr, TYPE_BOOL
        OptimizeNode clsExpr
        Match TokenEndOfInstruction
        ParseCodeBlock clsLoopCode
        Match TokenEnd
        Match TokenEndOfInstruction
        
        Set clsWhile.Expr = clsExpr
        Set clsWhile.Code = clsLoopCode
        clsCode.AddStatement clsWhile
    End If
End Sub


Private Sub ParseStatementReturn(ByVal clsCode As ParserCodeBlock)
    Dim clsExpr As ExpressionNode
    Dim clsRet As ParserStatementReturn
    
    If Match(TokenReturn) Then
        Set clsRet = New ParserStatementReturn
        If m_clsScanner.TokenID <> TokenEndOfInstruction Then
            ParseExpression clsExpr, False
            CheckTypes clsExpr, m_fncCurrent.DataType
            OptimizeNode clsExpr
            Match TokenEndOfInstruction
        Else
            If m_fncCurrent.DataType <> TYPE_NONE Then
                AddError PARSER_INVALID_VALUE, strMsg:="Funktion muss Wert zurckgeben"
            End If
        End If
        
        Set clsRet.Expr = clsExpr
        clsCode.AddStatement clsRet
    End If
End Sub


Private Sub ParseStatementIf(ByVal clsCode As ParserCodeBlock)
    Dim clsExpr As ExpressionNode
    Dim clsCodeIf As ParserCodeBlock
    Dim clsCodeElse As ParserCodeBlock
    Dim clsIf As ParserStatementIf
        
    If Match(TokenIf) Then
        Set clsIf = New ParserStatementIf
        Set clsCodeIf = New ParserCodeBlock
        
        ParseExpression clsExpr, False
        CheckTypes clsExpr, TYPE_BOOL
        OptimizeNode clsExpr
        Match TokenThen
        Match TokenEndOfInstruction
        ParseCodeBlock clsCodeIf
        
        If Match(TokenElse, False) Then
            Match TokenEndOfInstruction
            Set clsCodeElse = New ParserCodeBlock
            ParseCodeBlock clsCodeElse
        End If
                
        Match TokenEnd
        Match TokenEndOfInstruction
        
        Set clsIf.IfExpr = clsExpr
        Set clsIf.IfCode = clsCodeIf
        Set clsIf.ElseCode = clsCodeElse
        clsCode.AddStatement clsIf
    End If
End Sub


Private Sub ParseVarDeclare(ByVal clsCode As ParserCodeBlock)
    Dim clsExpr As ExpressionNode
    Dim clsVar  As ParserStatementDim
    
    Set clsVar = New ParserStatementDim
    
    If Match(TokenDim) Then
        Dim udtVar As VarInfo
        
        ParseVar udtVar, False, True

        If Match(TokenOpAssign, False) Then
            If udtVar.IsArray Then
                AddError PARSER_TYPE_MISMATCH, strMsg:="Array kann nicht direkt initialisiert werden"
            End If
            ParseExpression clsExpr, False
            CheckTypes clsExpr, udtVar.DataType
            OptimizeNode clsExpr
        End If
        
        If Len(udtVar.Name) > 0 Then
            If Not NameRegistryAdd(udtVar.Name, NAME_VARIABLE, udtVar.DataType, False, False, Not udtVar.Size Is Nothing) Then
                AddError PARSER_NAME_AMBIGUOS, strMsg:=udtVar.Name & " zweideutig"
            End If
        End If
        
        clsVar.Name = udtVar.Name
        clsVar.DataType = udtVar.DataType Or IIf(udtVar.IsArray, TYPE_ARRAY, 0)
        clsVar.IsArray = udtVar.IsArray
        clsVar.IsGlobal = m_fncCurrent Is Nothing
        Set clsVar.Size = udtVar.Size
        Set clsVar.AssignExpr = clsExpr
        clsCode.AddStatement clsVar
        'clsCode.AddVariables clsVar
    End If
End Sub

Private Sub ParseVar( _
    var As VarInfo, _
    Optional ByVal blnAllowRef As Boolean = False, _
    Optional ByVal blnAllowArraySize As Boolean = False _
)

    Dim strName As String
    Dim strType As String
    Dim blnRef  As Boolean
    Dim udeType As DATA_TYPE
    Dim blnArr  As Boolean
    Dim clsExpr As ExpressionNode
    
    If blnAllowRef Then blnRef = Match(TokenRef, False)
    
    If m_clsScanner.TokenID = TokenName Then
        strName = m_clsScanner.Consume()
        
        If Match(TokenBracketLeft, False) Then
            blnArr = True
            If Not Match(TokenBracketRight, False) Then
                If blnAllowArraySize Then
                    ParseExpression clsExpr, False
                    CheckTypes clsExpr, TYPE_INT
                    OptimizeNode clsExpr
                End If
                Match TokenBracketRight
            End If
        End If
        
        Match TokenAs
        If m_clsScanner.TokenID = TokenName Then
            strType = m_clsScanner.Consume()
            udeType = ParseDataTypeStr(strType)
            If udeType = TYPE_UNKNOWN Then
                AddError PARSER_TYPE_UNKNOWN, strMsg:=strType & ": Unbekannter Typ"
            End If
        Else
            AddError PARSER_UNEXPECTED_TOKEN, strMsg:="Typ erwartet"
        End If
        
        With var
            .Name = strName
            .DataType = udeType Or IIf(blnArr, TYPE_ARRAY, 0)
            .IsArray = blnArr
            .Reference = blnRef
            Set .Size = clsExpr
        End With
    Else
        AddError PARSER_UNEXPECTED_TOKEN, strMsg:="Variablenname erwartet"
    End If
End Sub


Private Sub ParseExpression(a As ExpressionNode, ByVal blnOptimize As Boolean)
    ParseExpressionRec a
    If blnOptimize Then OptimizeNode a
End Sub

Private Sub ParseExpressionRec(a As ExpressionNode)
    Dim b As ExpressionNode
    
    ParseExpressionPrec a, 0
    Do While Match(TokenOpAssign, False)
        ParseExpression b, False
        If a.DataType <> TYPE_UNKNOWN Then CheckTypes b, a.DataType
        
        If a.NodeType <> NODE_VAL_VARIABLE Then
            AddError PARSER_VAR_EXPECTED, strMsg:="Zuweisung nur an Variablen mglich"
        End If
        
        Set a = CreateNode(NODE_OP_ASSIGN, TYPE_UNKNOWN, 0, Nothing, a, b)
        CheckTypes a
    Loop
End Sub

Private Sub ParseExpressionPrec(a As ExpressionNode, ByVal lngPrec As Long)
    Dim b As ExpressionNode
    Dim i As Long
    Dim blnDone As Boolean
    
    If lngPrec = m_lngExprLkLvs Then
        ParseExpressionFactor a
    Else
        ParseExpressionPrec a, lngPrec + 1
        
        With m_udtExprLk(lngPrec)
            Do
                blnDone = True
                For i = 0 To .Count - 1
                    If Match(.Items(i).TokenType, False) Then
                        ParseExpressionPrec b, lngPrec + 1
                        Set a = CreateNode(.Items(i).NodeType, _
                                           TYPE_UNKNOWN, 0, Nothing, a, b)
                        CheckTypes a
                        blnDone = False
                        Exit For
                    End If
                Next
            Loop While Not blnDone
        End With
    End If
End Sub

Private Sub ParseExpressionFactor(a As ExpressionNode)
    Dim blnSign As Boolean
    
    blnSign = Match(TokenOpSub, False)
    
    Select Case m_clsScanner.TokenID
        Case TokenBoolFalse:        Set a = CreateNode(NODE_VAL_FALSE, TYPE_BOOL, False, Nothing)
                                    m_clsScanner.Consume
        Case TokenBoolTrue:         Set a = CreateNode(NODE_VAL_TRUE, TYPE_BOOL, True, Nothing)
                                    m_clsScanner.Consume
        
        Case TokenValue:            ParseExpressionFactorValue a
        Case TokenName:             ParseExpressionFactorName a
        
        Case TokenParenthesisLeft:  m_clsScanner.Consume
                                    ParseExpression a, False
                                    Match TokenParenthesisRight
                                    
        Case Else:                  AddError PARSER_UNEXPECTED_TOKEN
                                    Set a = CreateNode(NODE_NOT_DEFINED, TYPE_UNKNOWN, 0, Nothing)
    End Select
    
    If blnSign Then
        If a.DataType = TYPE_BYTE Then CheckTypes a, TYPE_SHORT
        Set a = CreateNode(NODE_OP_NEG, a.DataType, 0, Nothing, a)
    End If
End Sub

Private Sub ParseExpressionFactorName(a As ExpressionNode)
    Dim strName     As String
    Dim hVar        As Long
    Dim udeDataType As DATA_TYPE
    Dim udeNameType As NAME_TYPE
    Dim blnIsRef    As Boolean
    Dim clsExpr     As ExpressionNode
    
    strName = m_clsScanner.Consume()
    hVar = NameRegistryGetNameHandle(strName)
    If hVar = -1 Then
        AddError PARSER_UNKNOWN_NAME, strMsg:=strName & ": Unbekannter Name"
        udeDataType = TYPE_UNKNOWN
        udeNameType = NAME_VARIABLE
    Else
        udeDataType = NameRegistryInfo(hVar).DataType
        udeNameType = NameRegistryInfo(hVar).NameType
        blnIsRef = NameRegistryInfo(hVar).IsRef
    End If
    
    If Match(TokenParenthesisLeft, False) Then
        If udeNameType <> NAME_FUNCTION Then AddError PARSER_INVALID_FUNCTION, strMsg:=strName & ": Ungltige Funktion"
        ParseExpressionFactorNameFnc a, strName, udeDataType
    Else
        If udeNameType <> NAME_VARIABLE Then AddError PARSER_VAR_EXPECTED, strMsg:=strName & ": Keine Variable"
        ParseExpressionFactorNameVar a, strName, udeDataType, blnIsRef
    End If
End Sub

Private Sub ParseExpressionFactorNameFnc(a As ExpressionNode, ByVal strName As String, ByVal udeType As DATA_TYPE)
    Set a = CreateNode(NODE_OP_CALL, udeType, strName, Nothing)
    
    If Not Match(TokenParenthesisRight, False) Then
        ParseArgumentList strName, a
        Match TokenParenthesisRight
    End If
    
    If a.SubNodeCount <> GetFunction(strName).ParamCount Then
        AddError PARSER_WRONG_ARGUMENTS, strMsg:=strName & ": Falsche Anahl an Argumenten"
    End If
End Sub

Private Sub ParseExpressionFactorNameVar( _
    a As ExpressionNode, _
    ByVal strName As String, _
    ByVal udeType As DATA_TYPE, _
    ByVal blnIsRef As Boolean _
)

    Dim clsExpr As ExpressionNode
    
    Set a = CreateNode(NODE_VAL_VARIABLE, udeType, strName, Nothing)
    
    If Match(TokenBracketLeft, False) Then
        If (a.DataType And TYPE_ARRAY) = 0 Then
            AddError PARSER_TYPE_MISMATCH, strMsg:=strName & " ist kein Array"
        End If
        
        If Not Match(TokenBracketRight, False) Then
            ParseExpression clsExpr, False
            CheckTypes clsExpr, TYPE_INT
            OptimizeNode clsExpr
            Match TokenBracketRight
            
            ' a.ArrayIndex signalisiert jetzt Array,
            ' DataType muss nicht mehr darauf hinweisen
            Set a.ArrayIndex = clsExpr
            a.DataType = a.DataType And (Not TYPE_ARRAY)
        End If
    End If
    
    a.IsReference = blnIsRef
End Sub

Private Sub ParseArgumentList(ByVal strName As String, a As ExpressionNode)
    Dim i As Long
    Dim udtFnc As FunctionHead

    udtFnc = GetFunction(strName)
    
    i = 0
    Dim b As ExpressionNode
    Do
        ParseExpression b, False
        If i < udtFnc.ParamCount Then
            CheckTypes b, udtFnc.Params(i).DataType
            OptimizeNode b
            If udtFnc.Params(i).Reference Then
                If b.NodeType <> NODE_VAL_VARIABLE Then
                    AddError PARSER_TYPE_MISMATCH, strMsg:=strName & ": Argument " & (i + 1) & " kann nicht referenziert werden"
                Else
                    If NameRegistryInfo(NameRegistryGetNameHandle(b.Identifier)).IsFixedArray Then
                        AddError PARSER_TYPE_MISMATCH, strMsg:="Fixed Arrays knnen nicht referenziert bergeben werden"
                    End If
                End If
            End If
        Else
            AddError PARSER_WRONG_ARGUMENTS, strMsg:="Parameter " & (i + 1) & " ungltig"
        End If
        a.AddSubNode b
        i = i + 1
    Loop While Match(TokenSeperator, False)
End Sub

Private Sub ParseExpressionFactorValue(a As ExpressionNode)
    Dim strVal  As String
    
    strVal = m_clsScanner.Consume()
    If InStr(strVal, ".") > 0 Then
        Dim sngVal As Single
        If ParseFloat(strVal, sngVal) Then
            Set a = CreateNode(NODE_VAL_VALUE, TYPE_FLOAT, sngVal, Nothing)
        Else
            Set a = CreateNode(NODE_VAL_VALUE, TYPE_FLOAT, 0, Nothing)
            AddError PARSER_INVALID_VALUE, strMsg:="Ungltiger Wert: " & strVal
        End If
    Else
        Dim lngVal As Long
        If ParseLong(strVal, lngVal) Then
            If lngVal >= 0 And lngVal <= 255 Then
                Set a = CreateNode(NODE_VAL_VALUE, TYPE_BYTE, lngVal, Nothing)
            ElseIf lngVal >= -32768 And lngVal <= 32767 Then
                Set a = CreateNode(NODE_VAL_VALUE, TYPE_SHORT, lngVal, Nothing)
            Else
                Set a = CreateNode(NODE_VAL_VALUE, TYPE_INT, lngVal, Nothing)
            End If
        Else
            Set a = CreateNode(NODE_VAL_VALUE, TYPE_INT, 0, Nothing)
            AddError PARSER_INVALID_VALUE, strMsg:="Ungltiger Wert: " & strVal
        End If
    End If
End Sub

Private Function ParseFloat( _
    ByVal strVal As String, _
    ByRef sngVal As Single _
) As Boolean

    On Error GoTo ErrorHandler
    sngVal = CSng(Val(strVal))
    ParseFloat = True
ErrorHandler:
End Function

Private Function ParseLong( _
    ByVal strVal As String, _
    ByRef lngVal As Long _
) As Boolean

    On Error GoTo ErrorHandler
    lngVal = CLng(strVal)
    ParseLong = True
ErrorHandler:
End Function


Private Sub CheckTypes( _
    ByRef nd As ExpressionNode, _
    Optional ByVal dtResult As DATA_TYPE = TYPE_UNKNOWN _
)

    Dim dtOp    As DATA_TYPE
    Dim dtLeft  As DATA_TYPE
    Dim dtRight As DATA_TYPE
    Dim ndCheck As ExpressionNode
    Dim lngOpImpl As Long
    Dim lngNdLeftImpl As Long
    Dim lngNdRightImpl As Long
    
    If dtResult = TYPE_UNKNOWN Then
        dtOp = GetOpTypes(nd.NodeType)
        If nd.SubNodeCount = 0 Then Exit Sub
        dtLeft = nd.SubNode(0).DataType
        dtRight = nd.SubNode(1).DataType
        lngNdLeftImpl = GetImplicitConvPrec(dtLeft)
        lngNdRightImpl = GetImplicitConvPrec(dtRight)

        If (dtLeft And TYPE_ARRAY) And (dtRight And TYPE_ARRAY) Then
            If nd.NodeType <> NODE_OP_ASSIGN Then
                AddError PARSER_TYPE_MISMATCH, strMsg:="Arrays knnen nur komplett kopiert werden"
                Exit Sub
            End If
        ElseIf (dtLeft And TYPE_ARRAY) Or (dtRight And TYPE_ARRAY) Then
            If dtLeft <> -1 And dtRight <> -1 Then
                AddError PARSER_TYPE_MISMATCH, strMsg:="An ein Array kann nur ein anderes Array zugewiesen weden"
                Exit Sub
            End If
        End If

        If lngNdLeftImpl > lngNdRightImpl Then
            dtOp = dtOp And dtLeft
            Set ndCheck = nd.SubNode(1)
        Else
            dtOp = dtOp And dtRight
            Set ndCheck = nd.SubNode(0)
        End If
        
        If dtOp = 0 Then
            AddError PARSER_TYPE_MISMATCH
            Exit Sub
        End If
        
        lngOpImpl = GetImplicitConvPrec(dtOp)
        nd.DataType = dtOp
        
        If dtLeft <> dtRight Then
            Set ndCheck = CreateConversionNode(ndCheck.DataType, dtOp, ndCheck)
            If ndCheck.NodeType = NODE_NOT_DEFINED Then
                AddError PARSER_TYPE_MISMATCH
            End If
            
            If lngNdLeftImpl > lngNdRightImpl Then
                Set nd.SubNode(1) = ndCheck
            Else
                Set nd.SubNode(0) = ndCheck
            End If
        End If
        
        Select Case nd.NodeType
            Case NODE_OP_CMP_EQU, NODE_OP_CMP_GRT, NODE_OP_CMP_GRTEQU, _
                 NODE_OP_CMP_NEQU, NODE_OP_CMP_SML, NODE_OP_CMP_SMLEQU:
                    nd.DataType = TYPE_BOOL
        End Select
    Else
        If (nd.DataType And TYPE_ARRAY) Then
            If (dtResult And TYPE_ARRAY) Then
                If (nd.DataType And (Not TYPE_ARRAY)) <> (dtResult And (Not TYPE_ARRAY)) Then
                    AddError PARSER_TYPE_MISMATCH, strMsg:="Zu kopierendes Array muss gleichen Typen haben"
                End If
            End If
        End If
        
        lngOpImpl = GetImplicitConvPrec(nd.DataType)
        lngNdLeftImpl = GetImplicitConvPrec(dtResult)
        If lngNdLeftImpl > lngOpImpl Then
            Set nd = CreateConversionNode(nd.DataType, dtResult, nd)
            If nd.NodeType = NODE_NOT_DEFINED Then
                AddError PARSER_TYPE_MISMATCH
            End If
        ElseIf lngNdLeftImpl < lngOpImpl Then
            AddError PARSER_TYPE_MISMATCH
        End If
    End If
End Sub

Private Function CreateConversionNode( _
    ByVal dtFrom As DATA_TYPE, _
    ByVal dtTo As DATA_TYPE, _
    ByVal nd As ExpressionNode _
) As ExpressionNode

    Dim nt As NODE_TYPE: nt = NODE_NOT_DEFINED
    
    Select Case dtFrom
        Case TYPE_BYTE:
            Select Case dtTo
                Case TYPE_SHORT:    nt = NODE_CONV_BYTE_TO_SHORT
                Case TYPE_INT:      nt = NODE_CONV_BYTE_TO_INT
                Case TYPE_FLOAT:    nt = NODE_CONV_BYTE_TO_FLOAT
            End Select
        Case TYPE_SHORT:
            Select Case dtTo
                Case TYPE_INT:      nt = NODE_CONV_SHORT_TO_INT
                Case TYPE_FLOAT:    nt = NODE_CONV_SHORT_TO_FLOAT
            End Select
        Case TYPE_INT:
            Select Case dtTo
                Case TYPE_FLOAT:    nt = NODE_CONV_INT_TO_FLOAT
            End Select
    End Select
    
    Set CreateConversionNode = CreateNode(nt, dtTo, 0, Nothing, nd)
End Function


Private Function Match( _
    ByVal tk As TokenType, _
    Optional ByVal blnError As Boolean = True _
) As Boolean

    If m_clsScanner.Match(tk) Then
        Match = True
    Else
        If blnError Then
            AddError PARSER_UNEXPECTED_TOKEN, _
                     strMsg:="Erwartet: " & m_clsScanner.TokenTypeToString(tk)
        End If
    End If
End Function

Private Function CreateNode( _
    ByVal nt As NODE_TYPE, _
    ByVal dt As DATA_TYPE, _
    ByVal vl As Variant, _
    ByVal clsArrIdx As ExpressionNode, _
    ParamArray subs() _
) As ExpressionNode

    Dim cls As ExpressionNode
    Set cls = New ExpressionNode
    
    With cls
        .NodeType = nt
        .DataType = dt
        
        Set .ArrayIndex = clsArrIdx
        
        If nt = NODE_VAL_VARIABLE Or nt = NODE_OP_CALL Then
            .Identifier = CStr(vl)
        ElseIf nt = NODE_VAL_VALUE Or nt = NODE_VAL_FALSE Or nt = NODE_VAL_TRUE Then
            .Value = vl
        End If
        
        Dim i As Long
        For i = 0 To UBound(subs)
            .AddSubNode subs(i)
            Set .SubNode(i).Parent = cls
        Next
    End With
    
    Set CreateNode = cls
End Function


'#################################################
'#################################################
' OPERATOREN REIHENFOLGE


' Reihenfolge, in der Operatoren ausgewertet werden.
' Z.B. stehen * und / unter + und -, damit so
' Punkt-Vor-Strich-Rechnung beachtet wird.
' => Unterste Operatoren zuerst
Private Sub InitExpressionLookup()
    NewExprLookupLevel
    AddExprLookup TokenOpAnd, NODE_OP_LGC_AND
    AddExprLookup TokenOpOr, NODE_OP_LGC_OR
    NewExprLookupLevel
    AddExprLookup TokenOpEquals, NODE_OP_CMP_EQU
    AddExprLookup TokenOpGreater, NODE_OP_CMP_GRT
    AddExprLookup TokenOpGreaterEqu, NODE_OP_CMP_GRTEQU
    AddExprLookup TokenOpSmaller, NODE_OP_CMP_SML
    AddExprLookup TokenOpSmallerEqu, NODE_OP_CMP_SMLEQU
    AddExprLookup TokenOpUnequal, NODE_OP_CMP_NEQU
    NewExprLookupLevel
    AddExprLookup TokenOpBitAnd, NODE_OP_BAND
    AddExprLookup TokenOpBitOr, NODE_OP_BOR
    AddExprLookup TokenOpBitXor, NODE_OP_BXOR
    AddExprLookup TokenOpMod, NODE_OP_MOD
    NewExprLookupLevel
    AddExprLookup TokenOpAdd, NODE_OP_ADD
    AddExprLookup TokenOpSub, NODE_OP_SUB
    NewExprLookupLevel
    AddExprLookup TokenOpMul, NODE_OP_MUL
    AddExprLookup TokenOpDiv, NODE_OP_DIV
End Sub

Private Sub NewExprLookupLevel()
    ReDim Preserve m_udtExprLk(m_lngExprLkLvs)
    m_lngExprLkLvs = m_lngExprLkLvs + 1
End Sub

Private Sub AddExprLookup(ByVal tk As TokenType, ByVal nd As NODE_TYPE)
    With m_udtExprLk(m_lngExprLkLvs - 1)
        ReDim Preserve .Items(.Count)
        With .Items(.Count)
            .TokenType = tk
            .NodeType = nd
        End With
        .Count = .Count + 1
    End With
End Sub


'#################################################
'#################################################
' NAME REGISTRY


Private Function GetFunction(ByVal strName As String) As FunctionHead
    Dim i As Long
    For i = 0 To m_lngFncCount - 1
        If StrComp(strName, m_udtFunctions(i).Name, vbTextCompare) = 0 Then
            GetFunction = m_udtFunctions(i)
            Exit Function
        End If
    Next
End Function

Private Sub ClearRegistry()
    Erase m_udtRegNames
    m_lngRegLevels = 0
End Sub

Private Sub NameRegistryGrow()
    ReDim Preserve m_udtRegNames(m_lngRegLevels) As RegistryLevel
    m_lngRegLevels = m_lngRegLevels + 1
End Sub

Private Sub NameRegistryShrink()
    If m_lngRegLevels = 1 Then
        Erase m_udtRegNames
    Else
        ReDim Preserve m_udtRegNames(m_lngRegLevels - 2) As RegistryLevel
    End If
    
    m_lngRegLevels = m_lngRegLevels - 1
End Sub

Private Function NameRegistryAdd( _
    ByVal strName As String, _
    ByVal regType As NAME_TYPE, _
    ByVal dtaType As DATA_TYPE, _
    ByVal blnRef As Boolean, _
    ByVal blnParam As Boolean, _
    ByVal blnFixedArray As Boolean _
) As Boolean

    Dim i As Long
    
    If NameRegistryGetNameHandle(strName) = -1 Then
        With m_udtRegNames(m_lngRegLevels - 1)
            ReDim Preserve .Names(.NameCount) As RegistryName
            With .Names(.NameCount)
                .Name = strName
                .NameType = regType
                .DataType = dtaType
                .IsRef = blnRef
                .IsParam = blnParam
                .IsFixedArray = blnFixedArray
            End With
            .NameCount = .NameCount + 1
            
            NameRegistryAdd = True
        End With
    End If
End Function

Private Function NameRegistryInfo(ByVal hName As Long) As RegistryName
    Dim i As Long
    Dim j As Long
    
    i = hName And &HFFFF&
    j = hName \ &H10000
    
    NameRegistryInfo = m_udtRegNames(i).Names(j)
End Function

Private Function NameRegistryGetNameHandle(ByVal strName As String) As Long
    Dim i As Long
    Dim j As Long
    
    For i = 0 To m_lngRegLevels - 1
        With m_udtRegNames(i)
            For j = 0 To .NameCount - 1
                If StrComp(.Names(j).Name, strName, vbTextCompare) = 0 Then
                    NameRegistryGetNameHandle = i + j * &H10000
                    Exit Function
                End If
            Next
        End With
    Next
    
    NameRegistryGetNameHandle = -1
End Function


'#################################################
'#################################################
' OPERATOR TYPES


Private Sub InitOpTypes()
    Const ALL_NUM_TYPES = TYPE_INT Or TYPE_SHORT Or TYPE_BYTE Or TYPE_FLOAT
    Const ALL_TYPES = ALL_NUM_TYPES Or TYPE_BOOL
    
    ' Typen, die zu Operatoren kompatibel sind
    AddOpTypes NODE_OP_ADD, ALL_NUM_TYPES
    AddOpTypes NODE_OP_SUB, ALL_NUM_TYPES
    AddOpTypes NODE_OP_MUL, ALL_NUM_TYPES
    AddOpTypes NODE_OP_DIV, ALL_NUM_TYPES
    AddOpTypes NODE_OP_BAND, ALL_NUM_TYPES And (Not TYPE_FLOAT)
    AddOpTypes NODE_OP_BOR, ALL_NUM_TYPES And (Not TYPE_FLOAT)
    AddOpTypes NODE_OP_BXOR, ALL_NUM_TYPES And (Not TYPE_FLOAT)
    AddOpTypes NODE_OP_MOD, ALL_NUM_TYPES And (Not TYPE_FLOAT)
    AddOpTypes NODE_OP_ASSIGN, ALL_TYPES
    AddOpTypes NODE_OP_CMP_EQU, ALL_TYPES
    AddOpTypes NODE_OP_CMP_GRT, ALL_NUM_TYPES
    AddOpTypes NODE_OP_CMP_GRTEQU, ALL_NUM_TYPES
    AddOpTypes NODE_OP_CMP_NEQU, ALL_TYPES
    AddOpTypes NODE_OP_CMP_SML, ALL_NUM_TYPES
    AddOpTypes NODE_OP_CMP_SMLEQU, ALL_NUM_TYPES
    AddOpTypes NODE_OP_LGC_AND, TYPE_BOOL
    AddOpTypes NODE_OP_LGC_OR, TYPE_BOOL
    AddOpTypes NODE_OP_NEG, ALL_NUM_TYPES And (Not TYPE_BYTE)
    
    ' Reihenfolge fr Implizite Konvertierung
    ' Erster hat niedrigste Przedenz.
    ' Bool ist nicht dabei, sollte nicht in eine
    ' Zahl konvertiert werden drfen,
    ' genausowenig wie eine Zahl in einen Bool.
    AddToImplicitConvTable TYPE_BYTE
    AddToImplicitConvTable TYPE_SHORT
    AddToImplicitConvTable TYPE_INT
    AddToImplicitConvTable TYPE_FLOAT
End Sub

Private Function GetOpTypes(ByVal Op As NODE_TYPE) As DATA_TYPE
    Dim i As Long
    For i = 0 To m_lngOpTypCnt - 1
        If Op = m_udtOpTypes(i).Op Then
            GetOpTypes = m_udtOpTypes(i).Types
            Exit Function
        End If
    Next
    GetOpTypes = TYPE_UNKNOWN
End Function

Private Sub AddOpTypes(ByVal Op As NODE_TYPE, ByVal udeTypes As DATA_TYPE)
    ReDim Preserve m_udtOpTypes(m_lngOpTypCnt) As OperatorTypes
    m_udtOpTypes(m_lngOpTypCnt).Op = Op
    m_udtOpTypes(m_lngOpTypCnt).Types = udeTypes
    m_lngOpTypCnt = m_lngOpTypCnt + 1
End Sub

Private Function GetImplicitConvPrec(ByVal dt As DATA_TYPE) As Long
    Dim i As Long
    For i = 0 To m_lngImplConvCnt - 1
        If m_udeImplConv(i) = dt Then
            GetImplicitConvPrec = i
            Exit Function
        End If
    Next
    GetImplicitConvPrec = -1
End Function

Private Sub AddToImplicitConvTable(ByVal dt As DATA_TYPE)
    ReDim Preserve m_udeImplConv(m_lngImplConvCnt) As DATA_TYPE
    m_udeImplConv(m_lngImplConvCnt) = dt
    m_lngImplConvCnt = m_lngImplConvCnt + 1
End Sub


'#################################################
'#################################################
' ERROR HANDLING


Private Sub ClearErrors()
    Set m_colErrors = New Collection
End Sub

Public Property Get Errors() As Collection
    Set Errors = m_colErrors
End Property

Private Sub AddError( _
    ByVal udeErr As PARSER_ERROR_TYPE, _
    Optional ByVal lngLine As Long = -1, _
    Optional ByVal lngPos As Long = -1, _
    Optional ByVal strMsg As String = "" _
)

    Dim clsErr As ParserError
    Set clsErr = New ParserError
    
    clsErr.Init IIf(lngLine = -1, m_clsScanner.TokenLine, lngLine), _
                IIf(lngPos = -1, m_clsScanner.TokenPosition, lngPos), _
                udeErr, strMsg
    
    m_colErrors.Add clsErr
End Sub
