VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "StreamOGG"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Option Explicit


' OGG Vorbis Decoder
'
' Sehr einfach und kurz gehalten dank Vorbis SDK Bibliotheken,
' Code also hauptschlich im Unterverzeichnis OggVorbis.
' Hat einen groen Haufen bersetzen erspart.
'
' Arne Elster 2008/09


Implements IPlugin
Implements IPluginStream


Private Const MyDescription As String = "OggVorbis Decoder"
Private Const MyExtensions As String = "ogg"


Private Declare Sub CpyMem Lib "kernel32" Alias "RtlMoveMemory" ( _
    pDst As Any, pSrc As Any, ByVal cb As Long _
)

Private Declare Function vo_file_open Lib "OggVorbis.dll" ( _
    ByVal filename As String, ByVal mode As String _
) As Long

Private Declare Sub vo_file_close Lib "OggVorbis.dll" ( _
    ByVal handle As Long _
)

Private Declare Function vo_ov_open Lib "OggVorbis.dll" ( _
    ByVal filehandle As Long, VorbisFile As Any, _
    initial As Any, ByVal iBytes As Long _
) As Long

Private Declare Function vo_ov_clear Lib "OggVorbis.dll" ( _
    VorbisFile As Any _
) As Long

Private Declare Function vo_ov_info Lib "OggVorbis.dll" ( _
    VorbisFile As Any, ByVal link As Long _
) As Long

Private Declare Function vo_ov_time_total Lib "OggVorbis.dll" ( _
    VorbisFile As Any, ByVal link As Long _
) As Double

Private Declare Function vo_ov_bitrate Lib "OggVorbis.dll" ( _
    VorbisFile As Any, ByVal link As Long _
) As Long

Private Declare Function vo_ov_serialnumber Lib "OggVorbis.dll" ( _
    VorbisFile As Any, ByVal link As Long _
) As Long

Private Declare Function vo_ov_read Lib "OggVorbis.dll" ( _
    VorbisFile As Any, buffer As Any, _
    ByVal Length As Long, ByVal bigendianp As Long, _
    ByVal word As Long, ByVal sgned As Long, _
    ByRef bitstream As Long _
) As Long

Private Declare Function vo_ov_time_tell Lib "OggVorbis.dll" ( _
    VorbisFile As Any _
) As Double

Private Declare Function vo_ov_time_seek Lib "OggVorbis.dll" ( _
    VorbisFile As Any, ByVal s As Double _
) As Long

Private Declare Function vo_ov_comment Lib "OggVorbis.dll" ( _
    VorbisFile As Any, ByVal link As Long _
) As Long

Private Declare Function vo_write_tags_open Lib "OggVorbis.dll" ( _
    ByVal FileIn As String, ByVal FileOut As String _
) As Long

Private Declare Sub vo_write_tags_add Lib "OggVorbis.dll" ( _
    ByVal info As Long, ByVal name As String, ByVal value As String _
)

Private Declare Function vo_write_tags_close Lib "OggVorbis.dll" ( _
    ByVal info As Long _
) As Long

Private Type vorbis_tag
    name                    As String
    value                   As String
End Type

Private Type vorbis_comment
    user_comments           As Long
    comment_lengths         As Long
    Comments                As Long
    Vendor                  As Long
End Type

Private Type vorbis_info
    version                 As Long
    Channels                As Long
    rate                    As Long
    bitrate_upper           As Long
    bitrate_nominal         As Long
    bitrate_lower           As Long
    bitrate_window          As Long
    codec_setup             As Long
End Type

Private m_dblDuration       As Double
Private m_lngPosition       As Long
Private m_lngSamplerate     As Long
Private m_lngChannels       As Long
Private m_lngCurrSection    As Long
Private m_blnEOF            As Boolean

Private Const BUFFERSIZE    As Long = 4096
Private m_sngSamples(BUFFERSIZE - 1) As Single
Private m_lngSampleCount    As Long
Private m_lngSamplesUsed    As Long

Private m_strFile           As String
Private m_hFile             As Long
Private m_btVF(32& * 1024&) As Byte
Private m_udtInfo           As vorbis_info
Private m_clsInfo           As StreamTags

Private Property Get IPluginStream_Channels() As Long
    IPluginStream_Channels = m_udtInfo.Channels
End Property

Private Function IPluginStream_CreateNewStream() As IPluginStream
    Set IPluginStream_CreateNewStream = New StreamOGG
End Function

Private Property Get IPluginStream_Duration() As Long
    IPluginStream_Duration = CLng(m_dblDuration * 1000)
End Property

Private Property Get IPluginStream_Position() As Long
    IPluginStream_Position = CLng(vo_ov_time_tell(m_btVF(0)) * 1000)
End Property

Private Property Get IPluginStream_Samplerate() As Long
    IPluginStream_Samplerate = m_udtInfo.rate
End Property

Private Property Get IPluginStream_Description() As String
    IPluginStream_Description = MyDescription
End Property

Private Property Get IPluginStream_Extensions() As String()
    IPluginStream_Extensions = Split(MyExtensions, ";")
End Property

Private Function IPluginStream_StreamClose() As STREAM_RESULT
    If m_hFile = 0 Then Exit Function
    
    Dim i As Long
    Dim blnWriteTags As Boolean
    
    For i = 1 To m_clsInfo.TagCount
        If m_clsInfo.Tag(i).Changed Or m_clsInfo.Changed Then
            blnWriteTags = True
            Exit For
        End If
    Next
    
    vo_file_close m_hFile
    
    If blnWriteTags Then
        If WriteTags Then
            vo_file_close m_hFile
            Kill m_strFile
            Name m_strFile & ".tmp" As m_strFile
        End If
    End If
    
    vo_ov_clear m_btVF(0)
    Set m_clsInfo = Nothing
    m_hFile = 0
    m_blnEOF = False
    m_lngCurrSection = 0
    m_strFile = ""
End Function

Private Function WriteTags() As Boolean
    Dim hState As Long
    
    hState = vo_write_tags_open(m_strFile, m_strFile & ".tmp")
    If hState = 0 Then Exit Function
    
    Dim i As Long
    For i = 1 To m_clsInfo.TagCount
        With m_clsInfo.Tag(i)
            If Len(.Key) > 0 And Len(.value) > 0 Then
                vo_write_tags_add hState, UCase$(.Key), .value
            End If
        End With
    Next
    
    vo_write_tags_close hState
    WriteTags = FileLen(m_strFile & ".tmp") > 0
End Function

Private Property Get IPluginStream_StreamInfo() As StreamTags
    Set IPluginStream_StreamInfo = m_clsInfo
End Property

Private Function IPluginStream_StreamOpen(ByVal strPath As String) As STREAM_RESULT
    IPluginStream_StreamClose
    
    m_hFile = vo_file_open(strPath, "rb")
    If m_hFile = 0 Then
        IPluginStream_StreamOpen = RESULT_CANT_OPEN
        Exit Function
    End If
    m_strFile = strPath
    
    If vo_ov_open(m_hFile, m_btVF(0), ByVal 0&, 0) < 0 Then
        vo_file_close m_hFile: m_hFile = 0
        Exit Function
    End If
    
    Set m_clsInfo = New StreamTags
    
    Dim pUDT As Long
    pUDT = vo_ov_info(m_btVF(0), -1)
    CpyMem m_udtInfo, ByVal pUDT, Len(m_udtInfo)
    
    m_clsInfo.AddInfo "samplerate", m_udtInfo.rate & " Hz"
    m_clsInfo.AddInfo "channels", m_udtInfo.Channels
    
    m_dblDuration = vo_ov_time_total(m_btVF(0), -1)
    m_clsInfo.AddInfo "bitrate", CLng(vo_ov_bitrate(m_btVF(0), -1) / 1000) & " kbit/s"
    m_clsInfo.AddInfo "serialno", vo_ov_serialnumber(m_btVF(0), -1)
    m_clsInfo.AddInfo "version", m_udtInfo.version
    
    GetAllComments
    FillBuffer
    
    IPluginStream_StreamOpen = RESULT_SUCCESS
End Function

Private Function IPluginStream_StreamRead( _
    sngBuffer() As Single, _
    ByVal lngOffset As Long, _
    ByVal lngSamples As Long, _
    lngSamplesRead As Long _
) As STREAM_RESULT

    Dim lngNeeded As Long
    
    IPluginStream_StreamRead = RESULT_SUCCESS
    lngSamplesRead = 0
    
    If m_lngSampleCount = 0 Then
        If m_blnEOF Then
            IPluginStream_StreamRead = RESULT_END_OF_STREAM
        Else
            IPluginStream_StreamRead = RESULT_READ_ERROR
        End If
        Exit Function
    End If
    
    Do
        lngNeeded = Minimum(lngSamples, m_lngSampleCount - m_lngSamplesUsed)
        CpyMem sngBuffer(lngOffset + lngSamplesRead), m_sngSamples(m_lngSamplesUsed), 4 * lngNeeded
        lngSamples = lngSamples - lngNeeded
        m_lngSamplesUsed = m_lngSamplesUsed + lngNeeded
        lngSamplesRead = lngSamplesRead + lngNeeded
        
        If m_lngSamplesUsed = m_lngSampleCount Then
            Select Case FillBuffer()
                Case RESULT_SUCCESS
                    ' nichts tun
                Case RESULT_END_OF_STREAM
                    IPluginStream_StreamRead = RESULT_END_OF_STREAM
                Case Else
                    IPluginStream_StreamRead = RESULT_READ_ERROR
            End Select
        End If
    Loop While (lngSamples > 0) And (Not m_blnEOF)
End Function

Private Function FillBuffer() As STREAM_RESULT
    Const LITTLEENDIAN  As Long = 0
    Const WORDSIZE16BIT As Long = 2
    Const SIGNEDWORD    As Long = 1
    Dim ret             As Long
    Dim i               As Long
    Dim buffer(BUFFERSIZE \ 2 - 1) As Integer

    ret = vo_ov_read(m_btVF(0), buffer(0), BUFFERSIZE, _
                     LITTLEENDIAN, WORDSIZE16BIT, SIGNEDWORD, _
                     m_lngCurrSection)

    If ret >= 0 Then
        m_lngSampleCount = ret \ 2
        m_lngSamplesUsed = 0
        
        For i = 0 To m_lngSampleCount - 1
            m_sngSamples(i) = buffer(i) / 32767
        Next
    
        If ret = 0 Then
            m_blnEOF = True
            FillBuffer = RESULT_END_OF_STREAM
        Else
            FillBuffer = RESULT_SUCCESS
        End If
    Else
        FillBuffer = RESULT_UNKNOWN_ERROR
    End If
End Function

Private Function IPluginStream_StreamSeek(ByVal ms As Long) As STREAM_RESULT
    If vo_ov_time_seek(m_btVF(0), CDbl(ms / 1000)) = 0 Then
        IPluginStream_StreamSeek = RESULT_SUCCESS
    Else
        IPluginStream_StreamSeek = RESULT_UNKNOWN_ERROR
    End If
End Function

Private Sub GetAllComments()
    Dim udtComments         As vorbis_comment
    Dim pUDT                As Long
    Dim pComments()         As Long
    Dim pCommentLengths()   As Long
    Dim strComments()       As String
    Dim i                   As Long
    Dim j                   As Long
    Dim strTagName          As String
    Dim strTagValue         As String
    
    pUDT = vo_ov_comment(m_btVF(0), -1)
    CpyMem udtComments, ByVal pUDT, Len(udtComments)
    
    If udtComments.Vendor <> 0 Then
        m_clsInfo.AddInfo "vendor", PtrToAnsiZero(udtComments.Vendor)
    End If
    
    If udtComments.Comments > 0 Then
    
        ReDim pComments(udtComments.Comments - 1)
        ReDim pCommentLengths(udtComments.Comments - 1)
        ReDim strComments(udtComments.Comments - 1)
        
        CpyMem pCommentLengths(0), ByVal udtComments.comment_lengths, 4 * udtComments.Comments
        CpyMem pComments(0), ByVal udtComments.user_comments, 4 * udtComments.Comments
        
        For i = 0 To udtComments.Comments - 1
            strComments(i) = PtrToAnsi(pComments(i), pCommentLengths(i))
        Next
        
        For i = 0 To udtComments.Comments - 1
            strTagName = Left$(strComments(i), InStr(strComments(i), "=") - 1)
            strTagValue = Mid$(strComments(i), InStr(strComments(i), "=") + 1)
            
            For j = 1 To m_clsInfo.TagCount
                If StrComp(strTagName, m_clsInfo.Tag(j).Key, vbTextCompare) = 0 Then
                    GoTo NextComment
                End If
            Next
            
            If Len(strTagValue) > 0 Then
                m_clsInfo.AddTag LCase$(strTagName), strTagValue
            End If
NextComment:
        Next
    
    End If
End Sub

Private Function PtrToAnsiZero(ByVal ptr As Long) As String
    Dim strValue    As String
    Dim btVal       As Byte
    
    Do
        CpyMem btVal, ByVal ptr, 1
        ptr = ptr + 1
        If btVal = 0 Then Exit Do
        
        strValue = strValue & Chr(btVal)
    Loop
    
    PtrToAnsiZero = strValue
End Function

Private Function PtrToAnsi(ByVal ptr As Long, ByVal Length As Long) As String
    Dim btString() As Byte
    ReDim btString(Length - 1) As Byte
    
    CpyMem btString(0), ByVal ptr, Length
    
    PtrToAnsi = StrConv(btString, vbUnicode)
End Function

' #################################################
' #################################################

Private Property Get IPlugin_Description() As String
    IPlugin_Description = "Decodiert Ogg Vorbis Audio Dateien"
End Property

Private Property Get IPlugin_Name() As String
    IPlugin_Name = "Ogg Vorbis Stream"
End Property

Private Property Get IPlugin_PluginType() As MonoPluginLib.PLUGIN_TYPE
    IPlugin_PluginType = PLUGIN_STREAM
End Property
